import { Dataspace, Embedded, Reader, Ref, Schemas, Sturdy } from "@syndicate-lang/core";
import * as html from "@syndicate-lang/html";
import * as wsRelay from "@syndicate-lang/ws-relay";
import * as wakeDetector from './wake-detector.js';
import * as Shapes from './gen/shapes.js';

import {
    Engine,
    Mesh,
    Scene,
} from '@babylonjs/core/Legacy/legacy';

import { activeFloorMeshes, ShapeTree } from './shapes.js';
import { startEngine, CreatedScene } from './engine.js';

assertion type SceneHandle(ds: Embedded<Ref>);

function interpretScene(scene: Scene, sceneDs: Ref) {
    at sceneDs {
        during Shapes.Sprite({ "name": $name: string }) => spawn named `sprite:${name}` {
            console.log('+shape', name);
            on stop console.log('-shape', name);
            spriteMain(name, scene, sceneDs);
        }
    }
}

function spriteMain(name: string, scene: Scene, sceneDs: Ref) {
    at sceneDs {
        let currentShape = ShapeTree.empty(name, scene);
        on stop currentShape.remove();
        during Shapes.Sprite({ "name": name, "shape": $shape: Shapes.Shape }) => {
            console.log('=shape', name, shape);
            currentShape = currentShape.reconcile(name, shape);
        }
    }
}

function wsurl(): string {
    const scheme = (document.location.protocol.toLowerCase() === 'https:') ? 'wss' : 'ws';
    return `${scheme}://${document.location.host}/ws`;
}

function bootApp(ds: Ref, scene: Scene) {
    spawn named 'app' {
        at ds {
            const url = wsurl();
            const serverCap = Sturdy.asSturdyRef(new Reader<Ref>(
                '<ref "syndicate" [] #[pkgN9TBmEd3Q04grVG4Zdw==]>').next());
            const relayAddr = wsRelay.RelayAddress(Schemas.transportAddress.WebSocket(url));

            during wsRelay.Resolved({
                "addr": relayAddr,
                "sturdyref": serverCap,
                "resolved": $remoteDs_e: Embedded,
            }) => {
                const remoteDs = remoteDs_e.embeddedValue;

                on message wakeDetector.WakeEvent() => {
                    send message wsRelay.ForceRelayDisconnect(relayAddr);
                }

                at remoteDs {
                    during SceneHandle($sceneDs_e: Embedded) => {
                        const sceneDs = sceneDs_e.embeddedValue;
                        interpretScene(scene, sceneDs);
                    }
                }
            }
        }
    }
}

window.addEventListener('load', async () => {
    const scene = await startEngine(
        async (_canvas: HTMLCanvasElement, engine: Engine): Promise<CreatedScene> => ({
            scene: new Scene(engine),
            floorMeshes: () => activeFloorMeshes,
        }));
    Dataspace.boot(ds => {
        html.boot(ds);
        wsRelay.boot(ds, true);
        wakeDetector.boot(ds);
        bootApp(ds, scene);
    });
});
